#include <iostream>
#include <time.h>
#include <iomanip>
#include <fstream>
#include"graph.h"

#define CORRELATION

typedef Graph<double, double, double> GraphType;

typedef unsigned char uchar;

using namespace std;

//摜̐K
void Normalize(uchar *image, long width, long height)
{
	uchar max=1, *ptr;

	ptr = image;
	for(int i=width*height-1; i>=0; i--){
		if(max < *ptr){
			max = *ptr;
		}
		ptr++;
	}
	ptr = image;
	for(int i=0; i<width*height; i++){
		*(ptr++) = (uchar)(((double)*ptr / max) * 255);
	}
}

//PPM摜̓ǂݍ
uchar *readPPM(const char*filename, const long &width, const long &height)
{
	ifstream ifs(filename, ios::binary);
	uchar *image;
	char str[256];
	if(ifs == NULL){
		cout << "File Open Error" << endl;
		exit(1);
	}
	
	do{
		ifs.getline(str, 255);
	}while(str[0] == '#');
	if(strcmp(str, "P6") != 0){
		cout << "Not File P6" << endl;
		exit(1);
	}
	do{
		ifs.getline(str, 255);
	}while(str[0] == '#');
	sscanf_s(str, "%d%d", &width, &height);
	do{
		ifs.getline(str, 255);
	}while(str[0] == '#');

	image = new uchar[width*height*3];
	ifs.read((char*)image, width*height*3);
	ifs.close();

	return image;
}

//PGM摜̓ǂݍ
void writePGM(const char *filename, uchar *image, long width, long height)
{
	ofstream ofs(filename, ios::binary);
	if(ofs == NULL){
		cout << "File Open Error" << endl;
		exit(1);
	}

	ofs << "P5" << endl
		<< width << ' ' << height << endl
		<< 255 << endl;
	ofs.write((char*)image, width*height);
	ofs.close();
}

inline double max(double p, double q){
	return p > q? p : q;
}

//X[YRXg
inline double smooth_cost(uchar p, uchar q, int sobel)
{
	const double c0= 1, c1=50;
	double cost;
	
	cost = c0*exp(-sobel/c1)*abs(p-q);
	return cost;
}

//Vt^uEBhE
double ShiftableWindow(uchar *p, uchar *q, int lx, int rx, int ly, long width, long height)
{
	//EBhETCY
	const int mask = 6;
	//RXg
	const double c = 10;

	double c0, c1, c2, c3, c4;
	double first, second, temp;
	int p_ptr, q_ptr;
	const double sigma = 100.0;
	double cost;
	double gauss;
	int num = 0;
	double rLR=0, rL=0, rR=0, rL2=0, rR2=0;
	double gLR=0, gL=0, gR=0, gL2=0, gR2=0;
	double bLR=0, bL=0, bR=0, bL2=0, bR2=0;
	double denominator;
	double r_correlation, g_correlation, b_correlation;

	//c0̃RXgvZ
	c0 = 0;
	num=0;
	rLR=rL=rR=rL2=rR2=0;
	gLR=gL=gR=gL2=gR2=0;
	bLR=bL=bR=bL2=bR2=0;
	for(int y=-mask; y<=mask; y++){
		for(int x=-mask; x<=mask; x++){
			if(x+lx < 0 || x+lx >= width) continue;
			if(x+rx < 0 || x+rx >= width) continue;
			if(y+ly < 0 || y+ly >= height) continue;
			p_ptr = 3*(lx + ly*width) + 3*(x + y*width);
			q_ptr = 3*(rx + ly*width) + 3*(x + y*width);
			gauss = 1;//exp(-(x*x+y*y)/(2.0*sigma));
#ifdef SAD
			c0 += abs(p[p_ptr+0] - q[q_ptr+0]);
			c0 += abs(p[p_ptr+1] - q[q_ptr+1]);
			c0 += abs(p[p_ptr+2] - q[q_ptr+2]);
			num+=3;
#endif
#ifdef SSD
			c0 += pow(p[p_ptr+0] - q[q_ptr+0], 2);
			c0 += pow(p[p_ptr+1] - q[q_ptr+1], 2);
			c0 += pow(p[p_ptr+2] - q[q_ptr+2], 2);
			num+=3;
#endif 
#ifdef CORRELATION
			rLR += gauss*gauss*(p[p_ptr+0] * q[q_ptr+0]);
			gLR += gauss*gauss*(p[p_ptr+1] * q[q_ptr+1]);
			bLR += gauss*gauss*(p[p_ptr+2] * q[q_ptr+2]);

			rL += gauss*p[p_ptr+0];
			gL += gauss*p[p_ptr+1];
			bL += gauss*p[p_ptr+2];

			rL2 += gauss*gauss*(p[p_ptr+0] * p[p_ptr+0]);
			gL2 += gauss*gauss*(p[p_ptr+1] * p[p_ptr+1]);
			bL2 += gauss*gauss*(p[p_ptr+2] * p[p_ptr+2]);

			rR += gauss*(q[q_ptr+0]);
			gR += gauss*(q[q_ptr+1]);
			bR += gauss*(q[q_ptr+2]);

			rR2 += gauss*gauss*(q[q_ptr+0] * q[q_ptr+0]);
			gR2 += gauss*gauss*(q[q_ptr+1] * q[q_ptr+1]);
			bR2 += gauss*gauss*(q[q_ptr+2] * q[q_ptr+2]);
			num++;
#endif
		}
	}
#ifdef CORRELATION
	denominator = sqrt((num*rL2 - rL*rL)*(num*rR2 - rR*rR))+1;
	r_correlation = (num*rLR - rL*rR)/denominator;
	denominator = sqrt((num*gL2 - gL*gL)*(num*gR2 - gR*gR))+1;
	g_correlation =  (num*gLR - gL*gR)/denominator;
	denominator = sqrt((num*bL2 - bL*bL)*(num*bR2 - bR*bR))+1;
	b_correlation = (num*bLR - bL*bR)/denominator;
	c0 = 1.0 - (r_correlation + g_correlation + b_correlation)/3.0;
#else
	c0 /= num;
#endif


	//c1̃RXgvZ
	c1=0;
	num=0;
	rLR=rL=rR=rL2=rR2=0;
	gLR=gL=gR=gL2=gR2=0;
	bLR=bL=bR=bL2=bR2=0;
	for(int y=-mask*2; y<=0; y++){
		for(int x=-mask*2; x<=0; x++){
			if(x+lx < 0 || x+lx >= width) continue;
			if(x+rx < 0 || x+rx >= width) continue;
			if(y+ly < 0 || y+ly >= height) continue;
			p_ptr = 3*(lx + ly*width) + 3*(x + y*width);
			q_ptr = 3*(rx + ly*width) + 3*(x + y*width);
			gauss = 1;//(-exp(x*x+y*y)/(2.0*sigma));
#ifdef SAD
			c1 += abs(p[p_ptr+0] - q[q_ptr+0]);
			c1 += abs(p[p_ptr+1] - q[q_ptr+1]);
			c1 += abs(p[p_ptr+2] - q[q_ptr+2]);
			num+=3;
#endif
#ifdef SSD
			c1 += pow(p[p_ptr+0] - q[q_ptr+0], 2);
			c1 += pow(p[p_ptr+1] - q[q_ptr+1], 2);
			c1 += pow(p[p_ptr+2] - q[q_ptr+2], 2);
			num+=3;
#endif 
#ifdef CORRELATION
			rLR += gauss*gauss*(p[p_ptr+0] * q[q_ptr+0]);
			gLR += gauss*gauss*(p[p_ptr+1] * q[q_ptr+1]);
			bLR += gauss*gauss*(p[p_ptr+2] * q[q_ptr+2]);

			rL += gauss*p[p_ptr+0];
			gL += gauss*p[p_ptr+1];
			bL += gauss*p[p_ptr+2];

			rL2 += gauss*gauss*(p[p_ptr+0] * p[p_ptr+0]);
			gL2 += gauss*gauss*(p[p_ptr+1] * p[p_ptr+1]);
			bL2 += gauss*gauss*(p[p_ptr+2] * p[p_ptr+2]);

			rR += gauss*(q[q_ptr+0]);
			gR += gauss*(q[q_ptr+1]);
			bR += gauss*(q[q_ptr+2]);

			rR2 += gauss*gauss*(q[q_ptr+0] * q[q_ptr+0]);
			gR2 += gauss*gauss*(q[q_ptr+1] * q[q_ptr+1]);
			bR2 += gauss*gauss*(q[q_ptr+2] * q[q_ptr+2]);
			num++;
#endif
		}
	}
#ifdef CORRELATION
	denominator = sqrt((num*rL2 - rL*rL)*(num*rR2 - rR*rR))+1;
	r_correlation = (num*rLR - rL*rR)/denominator;
	denominator = sqrt((num*gL2 - gL*gL)*(num*gR2 - gR*gR))+1;
	g_correlation =  (num*gLR - gL*gR)/denominator;
	denominator = sqrt((num*bL2 - bL*bL)*(num*bR2 - bR*bR))+1;
	b_correlation = (num*bLR - bL*bR)/denominator;
	c1 = 1.0 - (r_correlation + g_correlation + b_correlation)/3.0;
#else
	c1 /= num;
#endif

	//c2̃RXgvZ
	c2=0;
	num=0;
	rLR=rL=rR=rL2=rR2=0;
	gLR=gL=gR=gL2=gR2=0;
	bLR=bL=bR=bL2=bR2=0;
	for(int y=-mask*2; y<=0; y++){
		for(int x=0; x<=mask*2; x++){
			if(x+lx < 0 || x+lx >= width) continue;
			if(x+rx < 0 || x+rx >= width) continue;
			if(y+ly < 0 || y+ly >= height) continue;
			p_ptr = 3*(lx + ly*width) + 3*(x + y*width);
			q_ptr = 3*(rx + ly*width) + 3*(x + y*width);
			gauss = 1;//exp(-(x*x+y*y)/(2.0*sigma));
#ifdef SAD
			c2 += abs(p[p_ptr+0] - q[q_ptr+0]);
			c2 += abs(p[p_ptr+1] - q[q_ptr+1]);
			c2 += abs(p[p_ptr+2] - q[q_ptr+2]);
			num+=3;
#endif
#ifdef SSD
			c2 += pow(p[p_ptr+0] - q[q_ptr+0], 2);
			c2 += pow(p[p_ptr+1] - q[q_ptr+1], 2);
			c2 += pow(p[p_ptr+2] - q[q_ptr+2], 2);
			num+=3;
#endif 
#ifdef CORRELATION
			rLR += gauss*gauss*(p[p_ptr+0] * q[q_ptr+0]);
			gLR += gauss*gauss*(p[p_ptr+1] * q[q_ptr+1]);
			bLR += gauss*gauss*(p[p_ptr+2] * q[q_ptr+2]);

			rL += gauss*p[p_ptr+0];
			gL += gauss*p[p_ptr+1];
			bL += gauss*p[p_ptr+2];

			rL2 += gauss*gauss*(p[p_ptr+0] * p[p_ptr+0]);
			gL2 += gauss*gauss*(p[p_ptr+1] * p[p_ptr+1]);
			bL2 += gauss*gauss*(p[p_ptr+2] * p[p_ptr+2]);

			rR += gauss*(q[q_ptr+0]);
			gR += gauss*(q[q_ptr+1]);
			bR += gauss*(q[q_ptr+2]);

			rR2 += gauss*gauss*(q[q_ptr+0] * q[q_ptr+0]);
			gR2 += gauss*gauss*(q[q_ptr+1] * q[q_ptr+1]);
			bR2 += gauss*gauss*(q[q_ptr+2] * q[q_ptr+2]);
			num++;
#endif
		}
	}
#ifdef CORRELATION
	denominator = sqrt((num*rL2 - rL*rL)*(num*rR2 - rR*rR))+1;
	r_correlation = (num*rLR - rL*rR)/denominator;
	denominator = sqrt((num*gL2 - gL*gL)*(num*gR2 - gR*gR))+1;
	g_correlation =  (num*gLR - gL*gR)/denominator;
	denominator = sqrt((num*bL2 - bL*bL)*(num*bR2 - bR*bR))+1;
	b_correlation = (num*bLR - bL*bR)/denominator;
	c2 = 1.0 - (r_correlation + g_correlation + b_correlation)/3.0;
#else
	c2 /= num;
#endif

	//c3̃RXgvZ
	c3=0;
	num=0;
	rLR=rL=rR=rL2=rR2=0;
	gLR=gL=gR=gL2=gR2=0;
	bLR=bL=bR=bL2=bR2=0;
	for(int y=0; y<=mask*2; y++){
		for(int x=-mask*2; x<=0; x++){
			if(x+lx < 0 || x+lx >= width) continue;
			if(x+rx < 0 || x+rx >= width) continue;
			if(y+ly < 0 || y+ly >= height) continue;
			p_ptr = 3*(lx + ly*width) + 3*(x + y*width);
			q_ptr = 3*(rx + ly*width) + 3*(x + y*width);
			gauss = 1;//exp(-(x*x+y*y)/(2.0*sigma));
#ifdef SAD
			c3 += abs(p[p_ptr+0] - q[q_ptr+0]);
			c3 += abs(p[p_ptr+1] - q[q_ptr+1]);
			c3 += abs(p[p_ptr+2] - q[q_ptr+2]);
			num+=3;
#endif
#ifdef SSD
			c3 += pow(p[p_ptr+0] - q[q_ptr+0], 2);
			c3 += pow(p[p_ptr+1] - q[q_ptr+1], 2);
			c3 += pow(p[p_ptr+2] - q[q_ptr+2], 2);
			num+=3;
#endif 
#ifdef CORRELATION
			rLR += gauss*gauss*(p[p_ptr+0] * q[q_ptr+0]);
			gLR += gauss*gauss*(p[p_ptr+1] * q[q_ptr+1]);
			bLR += gauss*gauss*(p[p_ptr+2] * q[q_ptr+2]);

			rL += gauss*p[p_ptr+0];
			gL += gauss*p[p_ptr+1];
			bL += gauss*p[p_ptr+2];

			rL2 += gauss*gauss*(p[p_ptr+0] * p[p_ptr+0]);
			gL2 += gauss*gauss*(p[p_ptr+1] * p[p_ptr+1]);
			bL2 += gauss*gauss*(p[p_ptr+2] * p[p_ptr+2]);

			rR += gauss*(q[q_ptr+0]);
			gR += gauss*(q[q_ptr+1]);
			bR += gauss*(q[q_ptr+2]);

			rR2 += gauss*gauss*(q[q_ptr+0] * q[q_ptr+0]);
			gR2 += gauss*gauss*(q[q_ptr+1] * q[q_ptr+1]);
			bR2 += gauss*gauss*(q[q_ptr+2] * q[q_ptr+2]);
			num++;
#endif
		}
	}
#ifdef CORRELATION
	denominator = sqrt((num*rL2 - rL*rL)*(num*rR2 - rR*rR))+1;
	r_correlation = (num*rLR - rL*rR)/denominator;
	denominator = sqrt((num*gL2 - gL*gL)*(num*gR2 - gR*gR))+1;
	g_correlation =  (num*gLR - gL*gR)/denominator;
	denominator = sqrt((num*bL2 - bL*bL)*(num*bR2 - bR*bR))+1;
	b_correlation = (num*bLR - bL*bR)/denominator;
	c3 = 1.0 - (r_correlation + g_correlation + b_correlation)/3.0;
#else
	c3 /= num;
#endif

	//c4̃RXgvZ
	c4=0;
	num=0;
	rLR=rL=rR=rL2=rR2=0;
	gLR=gL=gR=gL2=gR2=0;
	bLR=bL=bR=bL2=bR2=0;
	for(int y=0; y<=mask*2; y++){
		for(int x=0; x<=mask*2; x++){
			if(x+lx < 0 || x+lx >= width) continue;
			if(x+rx < 0 || x+rx >= width) continue;
			if(y+ly < 0 || y+ly >= height) continue;
			p_ptr = 3*(lx + ly*width) + 3*(x + y*width);
			q_ptr = 3*(rx + ly*width) + 3*(x + y*width);
			gauss = 1;//exp(-(x*x+y*y)/(2.0*sigma));
#ifdef SAD
			c4 += abs(p[p_ptr+0] - q[q_ptr+0]);
			c4 += abs(p[p_ptr+1] - q[q_ptr+1]);
			c4 += abs(p[p_ptr+2] - q[q_ptr+2]);
			num+=3;
#endif
#ifdef SSD
			c4 += pow(p[p_ptr+0] - q[q_ptr+0], 2);
			c4 += pow(p[p_ptr+1] - q[q_ptr+1], 2);
			c4 += pow(p[p_ptr+2] - q[q_ptr+2], 2);
			num+=3;
#endif 
#ifdef CORRELATION
			rLR += gauss*gauss*(p[p_ptr+0] * q[q_ptr+0]);
			gLR += gauss*gauss*(p[p_ptr+1] * q[q_ptr+1]);
			bLR += gauss*gauss*(p[p_ptr+2] * q[q_ptr+2]);

			rL += gauss*p[p_ptr+0];
			gL += gauss*p[p_ptr+1];
			bL += gauss*p[p_ptr+2];

			rL2 += gauss*gauss*(p[p_ptr+0] * p[p_ptr+0]);
			gL2 += gauss*gauss*(p[p_ptr+1] * p[p_ptr+1]);
			bL2 += gauss*gauss*(p[p_ptr+2] * p[p_ptr+2]);

			rR += gauss*(q[q_ptr+0]);
			gR += gauss*(q[q_ptr+1]);
			bR += gauss*(q[q_ptr+2]);

			rR2 += gauss*gauss*(q[q_ptr+0] * q[q_ptr+0]);
			gR2 += gauss*gauss*(q[q_ptr+1] * q[q_ptr+1]);
			bR2 += gauss*gauss*(q[q_ptr+2] * q[q_ptr+2]);
			num++;
#endif
		}
	}
#ifdef CORRELATION
	denominator = sqrt((num*rL2 - rL*rL)*(num*rR2 - rR*rR))+1;
	r_correlation = (num*rLR - rL*rR)/denominator;
	denominator = sqrt((num*gL2 - gL*gL)*(num*gR2 - gR*gR))+1;
	g_correlation =  (num*gLR - gL*gR)/denominator;
	denominator = sqrt((num*bL2 - bL*bL)*(num*bR2 - bR*bR))+1;
	b_correlation = (num*bLR - bL*bR)/denominator;
	c4 = 1.0 - (r_correlation + g_correlation + b_correlation)/3.0;
#else
	c4 /= num;
#endif 

	if(c1 < c2){
		first = c1;
		second = c2;
	}else{
		first = c2;
		second = c1;
	}
	
	if(second > c3){
		if(first > c3){
			temp = first;
			first = c3;
			second = temp;
		}else{
			second = c3;
		}
	}
	
	if(second > c4){
		if(first > c4){
			temp = first;
			first = c4;
			second = temp;
		}else{
			second = c4;
		}
	}
	cost = c*(c0+first+second);
	return cost;
}

//f[^RXg
inline double data_cost(uchar *p, uchar *q, int lx, int rx, int ly, long width, long height)
{
	return ShiftableWindow(p, q, lx, rx, ly, width, height);
	//}XNTCY
	const int mask_size = 5;
	//RXg
	const double c =60.0;
	const double sigma = 100.0;
	const double pi = 3.1415926535;
	double gauss;
	double cost = 0;
	int p_ptr, q_ptr;
	int num = 0;
	double rLR=0, rL=0, rR=0, rL2=0, rR2=0;
	double gLR=0, gL=0, gR=0, gL2=0, gR2=0;
	double bLR=0, bL=0, bR=0, bL2=0, bR2=0;
	double denominator;
	double r_correlation, g_correlation, b_correlation;

	for(int y=-mask_size; y <= mask_size; y++){
		for(int x=-mask_size; x <= mask_size; x++){
			if(x+lx < 0 || x+lx >= width) continue;
			if(x+rx < 0 || x+rx >= width) continue;
			if(y+ly < 0 || y+ly >= height) continue;
			p_ptr = 3*(lx + ly*width) + 3*(x + y*width);
			q_ptr = 3*(rx + ly*width) + 3*(x + y*width);
			gauss = exp(-(x*x+y*y)/(2.0*sigma));
#ifdef SAD
			cost += abs(p[p_ptr+0] - q[q_ptr+0]);
			cost += abs(p[p_ptr+1] - q[q_ptr+1]);
			cost += abs(p[p_ptr+2] - q[q_ptr+2]);
			num+=3;
#endif
#ifdef SSD
			cost += pow(p[p_ptr+0] - q[q_ptr+0], 2);
			cost += pow(p[p_ptr+1] - q[q_ptr+1], 2);
			cost += pow(p[p_ptr+2] - q[q_ptr+2], 2);
			num+=3;
#endif 
#ifdef CORRELATION
			rLR += gauss*gauss*(p[p_ptr+0] * q[q_ptr+0]);
			gLR += gauss*gauss*(p[p_ptr+1] * q[q_ptr+1]);
			bLR += gauss*gauss*(p[p_ptr+2] * q[q_ptr+2]);

			rL += gauss*p[p_ptr+0];
			gL += gauss*p[p_ptr+1];
			bL += gauss*p[p_ptr+2];

			rL2 += gauss*gauss*(p[p_ptr+0] * p[p_ptr+0]);
			gL2 += gauss*gauss*(p[p_ptr+1] * p[p_ptr+1]);
			bL2 += gauss*gauss*(p[p_ptr+2] * p[p_ptr+2]);

			rR += gauss*(q[q_ptr+0]);
			gR += gauss*(q[q_ptr+1]);
			bR += gauss*(q[q_ptr+2]);

			rR2 += gauss*gauss*(q[q_ptr+0] * q[q_ptr+0]);
			gR2 += gauss*gauss*(q[q_ptr+1] * q[q_ptr+1]);
			bR2 += gauss*gauss*(q[q_ptr+2] * q[q_ptr+2]);
			num++;
#endif
		}
	}

#ifdef CORRELATION
	
	denominator = sqrt((num*rL2 - rL*rL)*(num*rR2 - rR*rR))+1;
	r_correlation = (num*rLR - rL*rR)/denominator;

	denominator = sqrt((num*gL2 - gL*gL)*(num*gR2 - gR*gR))+1;
	g_correlation =  (num*gLR - gL*gR)/denominator;
	
	denominator = sqrt((num*bL2 - bL*bL)*(num*bR2 - bR*bR))+1;
	b_correlation = (num*bLR - bL*bR)/denominator;
	
	cost = 3.0 - (r_correlation + g_correlation + b_correlation);
	if(cost < 0){
		cost = 0;
	}
	num = 1;
#endif
	
	if(num == 0)
		return DBL_MAX;
	if(lx <64){
		return c*0.1*cost/num;
	}
	return c*cost/num;
}

//eKw̃lvZ
void HierarchicalAlpha(int *A, int size, int n)
{
	int N = 1 << n;
	int num;
	int *ptr = A;

	*(ptr++) = 0;
	*(ptr++) = N/2;

	for(int j=1; j<n; j++){
		num = (1 << (j-1));
		for(int k=0; k<num; k++){
			*ptr = ((1+4*k)*N) / (1 << (j+1));
			*(ptr+num) = ((3+4*k)*N) / (1 << (j+1));
			ptr++;
		}
		ptr += num;
	}
}

//\[xtB^̌vZ
//type == 0 E
//type == 1 ㉺
void SobelFilter(uchar *src, int *dest, long width, long height, int type)
{
	int *ptr;
	int a1, a2, a3, a4, a5, a6;

	ptr = dest;

	//E
	if(type == 0){
		for(int y=0; y<height; y++){
			for(int x=0; x<width-1; x++){
				if(y==0){
					a1 = a2 = 0;
				}else{
					//a1 = -1*src[3*(x  +(y-1)*width)+0]-1*src[3*(x  +(y-1)*width)+1]-1*src[3*(x  +(y-1)*width)+2];
					a1 = -1*(int)(0.299*src[3*(x  +(y-1)*width)+0]+0.587*src[3*(x  +(y-1)*width)+1]+0.114*src[3*(x  +(y-1)*width)+2]);
					//a2 =  1*src[3*(x+1+(y-1)*width)+0]-1*src[3*(x  +(y-1)*width)+1]-1*src[3*(x  +(y-1)*width)+2];
					a2 =  1*(int)(0.299*src[3*(x+1+(y-1)*width)+0]+0.587*src[3*(x  +(y-1)*width)+1]+0.114*src[3*(x  +(y-1)*width)+2]);
				}
				if(y==height-1){
					a3  = a4 = 0;
				}else{
					//a3 = -1*src[3*(x  +(y+1)*width)+0]-1*src[3*(x  +(y+1)*width)+1]-1*src[3*(x  +(y+1)*width)+2];
					a3 = -1*(int)(0.299*src[3*(x  +(y+1)*width)+0]+0.587*src[3*(x  +(y+1)*width)+1]+0.114*src[3*(x  +(y+1)*width)+2]);
					//a4 =  1*src[3*(x+1+(y+1)*width)+0]+1*src[3*(x  +(y+1)*width)+1]+1*src[3*(x  +(y+1)*width)+2];
					a4 =  1*(int)(0.299*src[3*(x+1+(y+1)*width)+0]+0.587*src[3*(x  +(y+1)*width)+1]+0.114*src[3*(x  +(y+1)*width)+2]);
				}
				//a5 = -2*src[3*(x  + y*width)+0]-2*src[3*(x  + y*width)+1]-2*src[3*(x  + y*width)+2];
				a5 = -2*(int)(0.299*src[3*(x  + y*width)+0]+0.587*src[3*(x  + y*width)+1]+0.114*src[3*(x  + y*width)+2]);
				//a6 =  2*src[3*(x+1+ y*width)+0]+2*src[3*(x+1+ y*width)+1]+2*src[3*(x+1+ y*width)+2];
				a6 =  2*(int)(0.299*src[3*(x+1+ y*width)+0]+0.587*src[3*(x+1+ y*width)+1]+0.114*src[3*(x+1+ y*width)+2]);
				*(ptr++) = abs(a1+a2+a3+a4+a5+a6);
			}
		}
		return;
	}

	//㉺
	if(type == 1){
		for(int y=0; y<height-1; y++){
			for(int x=0; x<width; x++){
				if(x == 0){
					a1 = a2 = 0;
				}else{
					//a1 = -1*src[3*(x-1+ y   *width)+0]-1*src[3*(x-1+ y   *width)+1]-1*src[3*(x-1+ y   *width)+2];
					a1 = -1*(int)(0.299*src[3*(x-1+ y   *width)+0]+0.587*src[3*(x-1+ y   *width)+1]+0.114*src[3*(x-1+ y   *width)+2]);
					//a2 =  1*src[3*(x-1+(y+1)*width)+0]+1*src[3*(x-1+(y+1)*width)+1]+1*src[3*(x-1+(y+1)*width)+2];
					a2 =  1*(int)(0.299*src[3*(x-1+(y+1)*width)+0]+0.587*src[3*(x-1+(y+1)*width)+1]+0.144*src[3*(x-1+(y+1)*width)+2]);
				}
				if(x == width-1){
					a3 = a4 = 0;
				}else{
					//a3 = -1*src[3*(x+1+ y   *width)+0]-1*src[3*(x+1+ y   *width)+1]-1*src[3*(x+1+ y   *width)+2];
					a3 = -1*(int)(0.299*src[3*(x+1+ y   *width)+0]+0.587*src[3*(x+1+ y   *width)+1]+0.114*src[3*(x+1+ y   *width)+2]);
					//a4 =  1*src[3*(x+1+(y+1)*width)+0]+1*src[3*(x+1+(y+1)*width)+1]+1*src[3*(x+1+(y+1)*width)+2];
					a4 =  1*(int)(0.299*src[3*(x+1+(y+1)*width)+0]+0.587*src[3*(x+1+(y+1)*width)+1]+0.114*src[3*(x+1+(y+1)*width)+2]);
				}

				//a5 = -2*src[3*(x+ y   *width)+0]-2*src[3*(x+ y   *width)+1]-2*src[3*(x+ y   *width)+2];
				a5 = -2*(int)(0.299*src[3*(x+ y   *width)+0]+0.587*src[3*(x+ y   *width)+1]+0.114*src[3*(x+ y   *width)+2]);
				//a6 =  2*src[3*(x+(y+1)*width)+0]+2*src[3*(x+(y+1)*width)+1]+2*src[3*(x+(y+1)*width)+2];
				a6 =  2*(int)(0.299*src[3*(x+(y+1)*width)+0]+0.587*src[3*(x+(y+1)*width)+1]+0.114*src[3*(x+(y+1)*width)+2]);
				*(ptr++) = abs(a1+a2+a3+a4+a5+a6);
			}
		}
		return;
	}
}

//XeIRXg
inline double StereoCost(uchar *p, uchar *q, int lx, int rx, int ly, long width, long height)
{

	const int mask_size = 10;
	const double sigma = 50.0;
	const double pi = 3.1415926535;
	double gauss;
	double cost = 0;
	int datasize = width*height*3;
	int p_ptr, q_ptr;
	int num = 0;
	double rLR=0, rL=0, rR=0, rL2=0, rR2=0;
	double gLR=0, gL=0, gR=0, gL2=0, gR2=0;
	double bLR=0, bL=0, bR=0, bL2=0, bR2=0;
	double denominator;
	double r_correlation, g_correlation, b_correlation;

	for(int y=-mask_size; y <= mask_size; y++){
		for(int x=-mask_size; x <= mask_size; x++){
			if(x+lx < 0 || x+lx >= width) continue;
			if(x+rx < 0 || x+rx >= width) continue;
			if(y+ly < 0 || y+ly >= height) continue;
			p_ptr = 3*(lx + ly*width) + 3*(x + y*width);
			q_ptr = 3*(rx + ly*width) + 3*(x + y*width);
			gauss = exp(-(x*x+y*y)/(2*sigma*sigma));
#ifdef SAD
			cost += abs(p[p_ptr+0] - q[q_ptr+0]);
			cost += abs(p[p_ptr+1] - q[q_ptr+1]);
			cost += abs(p[p_ptr+2] - q[q_ptr+2]);
			num+=3;
#endif
#ifdef SSD
			cost += pow(p[p_ptr+0] - q[q_ptr+0], 2);
			cost += pow(p[p_ptr+1] - q[q_ptr+1], 2);
			cost += pow(p[p_ptr+2] - q[q_ptr+2], 2);
			num+=3;
#endif 
#ifdef CORRELATION
			rLR += gauss*gauss*p[p_ptr+0] * q[q_ptr+0];
			gLR += gauss*gauss*p[p_ptr+1] * q[q_ptr+1];
			bLR += gauss*gauss*p[p_ptr+2] * q[q_ptr+2];

			rL += gauss*p[p_ptr+0];
			gL += gauss*p[p_ptr+1];
			bL += gauss*p[p_ptr+2];

			rL2 += gauss*gauss*p[p_ptr+0] * p[p_ptr+0];
			gL2 += gauss*gauss*p[p_ptr+1] * p[p_ptr+1];
			bL2 += gauss*gauss*p[p_ptr+2] * p[p_ptr+2];

			rR += gauss*q[q_ptr+0];
			gR += gauss*q[q_ptr+1];
			bR += gauss*q[q_ptr+2];

			rR2 += gauss*gauss*q[q_ptr+0] * q[q_ptr+0];
			gR2 += gauss*gauss*q[q_ptr+1] * q[q_ptr+1];
			bR2 += gauss*gauss*q[q_ptr+2] * q[q_ptr+2];
			num++;
#endif
		}
	}

#ifdef CORRELATION
	
	if((denominator = sqrt((num*rL2 - rL*rL)*(num*rR2 - rR*rR))) != 0){
		r_correlation = (num*rLR - rL*rR)/denominator;
	}else{
		r_correlation = 0;
	}
	if((denominator = sqrt((num*gL2 - gL*gL)*(num*gR2 - gR*gR))) != 0){
		g_correlation =  (num*gLR - gL*gR)/denominator;
	}else{
		g_correlation = 0;
	}
	if((denominator = sqrt((num*bL2 - bL*bL)*(num*bR2 - bR*bR))) != 0){
		b_correlation = (num*bLR - bL*bR)/denominator;
	}else{
		b_correlation = 0;
	}
	
	
	cost = (r_correlation + g_correlation + b_correlation);
	num = 1;
#endif

	if(num == 0)
		return 0;
	return cost/num;
}

//YCrCbϊ
void ConvertYCrCb(unsigned char*image, int width, int height)
{
	unsigned char r, g, b;

	for(int i=0; i<width*height; i++){
		r = image[3*i+0];
		g = image[3*i+1];
		b = image[3*i+2];

		image[3*i+0] = (unsigned char)(0.299*r + 0.587*g + 0.114*b);
		image[3*i+1] = (unsigned char)(0.500*r - 0.419*g - 0.081*b + 128);
		image[3*i+2] = (unsigned char)(-0.169*r - 0.332*g + 0.500*b + 128);
	}
}

int main()
{
	GraphType *g;
	const int N = 7*2;
	const int disparity_max = 1 << (N/2);
	int A[disparity_max], A_num, A_offset;
	int *alpha, min_alpha;
	int index, p, q;
	uchar *left_image, *right_image;
	int *sobel_vertical, *sobel_horizontal;
	uchar *disparity;
	double dis_cost, mdis_cost;
	long width, height, imagesize;
	long  vertical, horizontal;
	long start, offset;
	double source_cost, sink_cost;
	double flow, E=DBL_MAX;
	const double infinity = DBL_MAX;
	bool success=true;
	const int loop_max = 10;

	//XeI摜̓ǂݍ
	left_image = readPPM("left_image.ppm", width, height);
	right_image = readPPM("right_image.ppm", width, height);
	//ConvertYCrCb(left_image, width, height);
	//ConvertYCrCb(left_image, width, height);
	HierarchicalAlpha(A, 64, N/2);

	//em[ȟvZ
	imagesize = width*height;
	vertical = (width-1)*height;
	horizontal = width*(height-1);

	//m
	disparity = new uchar[imagesize];
	alpha = new int[imagesize];
	sobel_vertical = new int[vertical];
	sobel_horizontal = new int[horizontal];
	SobelFilter(left_image, sobel_vertical, width, height, 0);
	SobelFilter(left_image, sobel_horizontal, width, height, 1);
	unsigned char*hoge;
	hoge = new unsigned char[vertical];
	for(int i=0; i<vertical; i++){
		hoge[i] = sobel_vertical[i];
	}
	writePGM("vertical.pgm", hoge, width-1, height);
	delete [] hoge;

	hoge = new unsigned char[horizontal];
	for(int i=0; i<horizontal; i++){
		hoge[i] = sobel_horizontal[i];
	}
	writePGM("horizontal.pgm", hoge, width, height-1);
	delete [] hoge;


	//XeI
	for(int y=0; y<height; y++){
		cout << "Stereo: " << setw(3) << y*100/(height-1) << "\%\r";
		cout.flush();
		for(int x=0; x<width; x++){
			mdis_cost = -DBL_MAX;
			disparity[x+y*width] = 0;
			for(int dis=disparity_max; dis >= 0; dis--){
				if(x-dis < 0) continue;
				dis_cost = -ShiftableWindow(left_image, right_image, x, x-dis, y, width, height);
				if(mdis_cost <= dis_cost){
					mdis_cost = dis_cost;
					disparity[x+y*width] = dis;
				}
			}
			
		}
	}
	cout << endl;
	uchar *stereo;
	stereo = new uchar[width*height];
	for(int i=0; i<width*height; i++){
		stereo[i] = disparity[i];
	}
	Normalize(stereo, width, height);
	writePGM("stereo.pgm", stereo, width, height);
	delete [] stereo;

	clock_t start_clock, end_clock;
	start_clock = clock();
	//OtJbg
	for(int looper=0; looper<loop_max; looper++){
		success = true;
		A_offset = 0;
		for(int hierarchy=0; hierarchy<N; hierarchy++){
			cout << "loop " << setw(2) << looper << " hierarchy " << setw(2) << hierarchy+1 << "\r";
			cout.flush();
			//Kw̌vZ
			if(hierarchy % 2 == 0){
				if(hierarchy == 0){
					A_num = 1;
				}else{
					A_num = 1 << (int)((hierarchy-1)/2);
				}
			}
			g = new GraphType(imagesize+vertical+horizontal, (vertical+horizontal)*2);

			//m[h̒ǉ
			for(int i=imagesize+vertical+horizontal; i>=0; i--){
				g->add_node();
			}

			//m[h̃RXgݒ
			for(int y=0; y<height; y++){
				for(int x=0; x<width; x++){
					index = x+y*width;
					alpha[index] = A[A_offset];
					min_alpha = INT_MAX;
					//em[h̃lݒ@arg min |alpha - disparity|
					for(int k=1; k<A_num; k++){
						if(min_alpha > abs(A[A_offset+k] - disparity[index]) ){
							min_alpha  = abs(A[A_offset+k] - disparity[index]);
							alpha[index] = A[A_offset+k];
						}
					}
					//fl(x,y)-fr(x-,y)
					source_cost = data_cost(left_image, right_image, x, x-disparity[index], y, width, height);
					//fl(x,y)-fr(x-,y)
					sink_cost = data_cost(left_image, right_image, x, x-alpha[index], y, width, height);
					g->add_tweights(index, source_cost, sink_cost); 
				}
			}

			//gm[h̃RXgݒ
			start = imagesize;
			offset = 0;
			for(int y=0; y<height; y++){
				for(int x=0; x<width-1; x++){
					//@sink
					//      p--q
					//		source
					p = x+y*width;
					q = (x+1)+y*width;
					source_cost = smooth_cost(disparity[p], disparity[q], sobel_vertical[offset]);
					sink_cost = smooth_cost(alpha[p], alpha[q], sobel_vertical[offset]);
					g->add_tweights(start+offset, source_cost, sink_cost);
					//V(p,q) > V(p,q)
					if(source_cost > sink_cost){
						g->add_edge(start+offset, p, max(0, smooth_cost(disparity[p], alpha[q], sobel_vertical[offset]) - sink_cost), infinity);
						g->add_edge(start+offset, q, max(0, smooth_cost(alpha[p], disparity[q], sobel_vertical[offset]) - sink_cost), infinity);
					}
					//V(p,q) < V(p,q)
					else{
						g->add_edge(start+offset, p, infinity, max(0, smooth_cost(alpha[p], disparity[q], sobel_vertical[offset]) - source_cost));
						g->add_edge(start+offset, q, infinity, max(0, smooth_cost(disparity[p], alpha[q], sobel_vertical[offset]) - source_cost));
					}
					offset++;
				}
			}

			start = imagesize+vertical;
			offset = 0;
			for(int y=0; y<height-1; y++){
				for(int x=0; x<width; x++){
					//		 p
					//		sink--source
					//			 q
					p = x+y*width;
					q = x+(y+1)*width;
					source_cost = smooth_cost(disparity[p], disparity[q], sobel_vertical[offset]);
					sink_cost = smooth_cost(alpha[p], alpha[q], sobel_vertical[offset]);
					g->add_tweights(start+offset, source_cost, sink_cost);
					//V(p,q) > V(p,q)
					if(source_cost > sink_cost){
						g->add_edge(start+offset, p, max(0, smooth_cost(disparity[p], alpha[q], sobel_vertical[offset]) - sink_cost), infinity);
						g->add_edge(start+offset, q, max(0, smooth_cost(alpha[p], disparity[q], sobel_vertical[offset]) - sink_cost), infinity);
					}
					//V(p,q) < V(p,q)
					else{
						g->add_edge(start+offset, p, infinity, max(0, smooth_cost(alpha[p], disparity[q], sobel_vertical[offset]) - source_cost));
						g->add_edge(start+offset, q, infinity, max(0, smooth_cost(disparity[p], alpha[q], sobel_vertical[offset]) - source_cost));
					}
					offset++;
				}
			}

			//ŏJbg
			flow = g->maxflow();
			if(flow < E){
				E = flow;
				success = false;
				//xt
				for(int i=0; i<imagesize; i++){
					if(g->what_segment(i) == GraphType::SOURCE){
						disparity[i] = alpha[i];
					}
				}
			}

			A_offset += A_num;
			delete g;
		}
		if(success) break;
	}
	end_clock = clock();
	cout << end_clock - start_clock <<"[ms]" <<endl;
	Normalize(disparity, width, height);
	writePGM("result.pgm", disparity, width, height);
	delete [] left_image;
	delete [] right_image;
	delete [] disparity;
	delete [] alpha;
	
	delete [] sobel_vertical;
	delete [] sobel_horizontal;
	getchar();
}